/*
 * display.c
 *
 * Created: 2017-07-16 21:49:22
 *  Author: doom
 */ 

#define F_CPU 8000000
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <avr/pgmspace.h>
#include "display.h"


#define LCD_PORT		PORTB
#define LCD_DDR			DDRB
#define LCD_DB4			3
#define LCD_DB5			5
#define LCD_DB6			2
#define LCD_DB7			4
#define LCD_RS			7
#define LCD_EN			6
#define LCD_DATA_PORT	LCD_PORT
#define LCD_COM_PORT	LCD_PORT
#define LCD_DATA_DDR	LCD_DDR
#define LCD_COM_DDR		LCD_DDR

//---------------------------------------------------
// Komendy sterujce wywietlaczem
#define LCDC_CLS				0x01
#define LCDC_HOME				0x02
#define LCDC_MODE				0x04
	#define LCDC_MODER			0x02
	#define LCDC_MODEL			0
	#define LCDC_MODEMOVE		0x01
#define LCDC_ON 				0x08
	#define LCDC_ONDISPLAY		0x04
	#define LCDC_ONCURSOR		0x02
	#define LCDC_ONBLINK		0x01
#define LCDC_SHIFT				0x10
	#define LCDC_SHIFTDISP		0x08
	#define LCDC_SHIFTR			0x04
	#define LCDC_SHIFTL			0
#define LCDC_FUNC				0x20
	#define LCDC_FUNC8b			0x10
	#define LCDC_FUNC4b			0
	#define LCDC_FUNC2L			0x08
	#define LCDC_FUNC1L			0
	#define LCDC_FUNC5x10		0x4
	#define LCDC_FUNC5x7		0
#define LCDC_CGA				0x40
#define LCDC_DDA				0x80

//---------------------------------------------------
#define MAX_SPEC_CHARS			8
/*const char empty[8] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
const char degree[8] = {0x06, 0x09, 0x09, 0x06, 0x00, 0x00, 0x00, 0x00};
const char cdegree[8] = {0x00, 0x18, 0x1B, 0x04, 0x04, 0x04, 0x03, 0x00};
const char maxSymbol[8] = {0x1F, 0x04, 0x0E, 0x15, 0x04, 0x04, 0x04, 0x00};	
const char minSymbol[8] = {0x04, 0x04, 0x04, 0x15, 0x0E, 0x04, 0x1F, 0x00};		
*/

//------------------------------------------------------------------
static void LCDInit(void);
static void LCDCommand(uint8_t command);
static void LCDData(uint8_t data);
static void LCDClear(void);
static void LCDHome(void);
static void LCDLocate(uint8_t lcd_y, uint8_t lcd_x);
static void LCDSendHalf(uint8_t byte);
static void LCDSendByte(uint8_t data);
static void LCDSetSpecChar(uint8_t spec, uint8_t* pData);
static uint8_t *Value2String(int16_t val, uint8_t *dst, uint8_t len, uint8_t dot);

//------------------------------------------------------------------
void DispInit(void){	
	LCDInit();
}

//------------------------------------------------------------------
void DispPosition(uint8_t posY, uint8_t posX){
	LCDLocate(posY, posX);
}

//------------------------------------------------------------------
void DispText(uint8_t *src){
	while((src > 0) && (*src > 0)) {
		LCDData(*src);
		src++;
	}
}

//------------------------------------------------------------------
void DispMsg(uint8_t *src, uint8_t posY, uint8_t posX){
	DispPosition(posY, posX);
	DispText(src);
}

//------------------------------------------------------------------
void DispValue(int16_t val){
	
	uint8_t buff[4];
	DispText(Value2String((int16_t)val, buff, 3, 0));
}

//------------------------------------------------------------------
void DispSimpleValue(uint16_t val){
	uint8_t buff[3];
	uint8_t i;
	
	if (val > 999) {
		val = 999;
	}
	
	for(i = 0; i < 3; i++){	
		if ((val > 0) || (i == 0)){
			buff[i] = (((uint8_t)(val % 10)) + '0');
			val /= 10;
		} else {
			buff[i] = ' ';
		}
	}
	LCDData(buff[2]);
	LCDData(buff[1]);
	LCDData(buff[0]);
}


//------------------------------------------------------------------
static uint8_t *Value2String(int16_t val, uint8_t *dst, uint8_t len, uint8_t dot){
	uint8_t minus = 0;
	uint8_t status = 1;
	uint8_t i = 0;
	uint8_t d;
	uint8_t *offset;
	
	if (val < 0) {
		minus = 1;
		val *= -1;
	}
	
	while(status > 0){
		d = (uint8_t)(val % 10);
		val /= 10;
			
		*dst = (d + '0');
		dst++;
		i++;
			
		if (dot == i){
			*dst = '.';
			dst++;
			i++;
			
			if (val == 0){
				*dst = '0';
				dst++;
				i++;
			}
		}
		
		if (val == 0){
			if (dot >= i){
				status = 0;
			}
		}
	}
	
	if (minus > 0){
		*dst = '-';
		dst++;
		i++;
	}
	
	offset = (dst - 1);
	
	while (i < len){
		*dst = ' ';
		dst++;
		i++;
	}

	return offset;
}

//---------------------------------------------------
static void LCDInit(void){
	//inits ports
	LCD_DATA_PORT |= (1<<LCD_DB4 | 1<<LCD_DB5 | 1<<LCD_DB6 | 1<<LCD_DB7);
	LCD_COM_PORT  |= (1<<LCD_EN | 1<<LCD_RS);
	LCD_DATA_DDR  |= (1<<LCD_DB4 | 1<<LCD_DB5 | 1<<LCD_DB6 | 1<<LCD_DB7);
	LCD_COM_DDR   |= (1<<LCD_EN | 1<<LCD_RS);
	_delay_ms(10);
	LCD_COM_PORT &= ~(1<<LCD_RS);
	_delay_ms(20);
	LCDSendHalf((LCDC_FUNC|LCDC_FUNC8b) >> 0x04);
	_delay_ms(10);
	LCDSendHalf((LCDC_FUNC|LCDC_FUNC8b) >> 0x04);
	_delay_ms(2);
	LCDSendHalf((LCDC_FUNC|LCDC_FUNC8b) >> 0x04);
	_delay_ms(2);
	LCDSendHalf((LCDC_FUNC|LCDC_FUNC4b) >> 0x04);
	_delay_ms(2);

	LCDCommand(LCDC_FUNC|LCDC_FUNC4b|LCDC_FUNC2L|LCDC_FUNC5x7);
	LCDCommand(LCDC_ON);
	LCDClear();
	LCDCommand(LCDC_MODE|LCDC_MODER);
	LCDCommand(LCDC_ON|LCDC_ONDISPLAY);
	
	/*LCDSetSpecChar(0, empty);	
	LCDSetSpecChar(1, degree);
	LCDSetSpecChar(2, cdegree);
	LCDSetSpecChar(3, maxSymbol);
	LCDSetSpecChar(4, minSymbol);
	*/
}

//---------------------------------------------------
static void LCDSendHalf(uint8_t byte){
	LCD_COM_PORT |= (1<<LCD_EN);
	
	if (byte & (1<<0))
	LCD_DATA_PORT |= (1<<LCD_DB4);
	else
	LCD_DATA_PORT &= ~(1<<LCD_DB4);
	
	if (byte & (1<<1))
	LCD_DATA_PORT |= (1<<LCD_DB5);
	else
	LCD_DATA_PORT &= ~(1<<LCD_DB5);

	if (byte & (1<<2))
	LCD_DATA_PORT |= (1<<LCD_DB6);
	else
	LCD_DATA_PORT &= ~(1<<LCD_DB6);

	if (byte & (1<<3))
	LCD_DATA_PORT |= (1<<LCD_DB7);
	else
	LCD_DATA_PORT &= ~(1<<LCD_DB7);
	
	LCD_COM_PORT &= ~(1<<LCD_EN);
	_delay_ms(1);
}

//---------------------------------------------------
static void LCDSendByte(uint8_t data){
	LCDSendHalf(data>>0x04);
	LCDSendHalf(data);
	//for the lowest possible power consumption:
	LCD_DATA_PORT |= (1<<LCD_EN | 1<<LCD_RS | 1<<LCD_DB4 | 1<<LCD_DB5 | 1<<LCD_DB6 | 1<<LCD_DB7);
	//DelayMs(1);
}

//---------------------------------------------------
static void LCDCommand(uint8_t command){
	LCD_COM_PORT &= ~(1<<LCD_RS);
	LCDSendByte(command);
}

//---------------------------------------------------
static void LCDData(uint8_t data){
	LCD_COM_PORT |= (1<<LCD_RS);
	LCDSendByte(data);
}

//---------------------------------------------------
static void LCDClear(void){
	LCDCommand(LCDC_CLS);
	_delay_ms(2);
}

//---------------------------------------------------
static void LCDHome(void){
	LCDCommand(LCDC_HOME);
}

//---------------------------------------------------
static void LCDLocate(uint8_t lcd_y, uint8_t lcd_x){
	LCDCommand((lcd_y*0x40+lcd_x)|0x80);
}

//---------------------------------------------------
static void LCDSetSpecChar(uint8_t spec, uint8_t *pData){
	char count;
	if (spec < (MAX_SPEC_CHARS - 1)){
		spec *= 8;
		LCDCommand(LCDC_CGA | spec);
		count = 0;
		while (count < 8){
			LCDData(*pData);
			pData++;
			count++;
		}
	}
}
